// descriptor.cpp -- main program, responsible for reading in the file (by
// calling the appropriate file reader), preprocessing the file appropriately,
// call all extraction functions in turn and finally output the extracted
// descriptors to standard output.

#include <stdio.h>
#include <math.h>

#include "common.h"
#include "centroid.h"
/* #include "wavelet.h" */
#include "mfcc.h"
#include "f1cc.h"

#include "vorbis.h"
#include "mp3.h"
#include "flac.h"
#include "wave.h"

void remove_dc_offset(Sample &s)
{
	double sum_x = 0.0;
	size_t N = std::min(s.samples.size(), (size_t)FEATURE_SAMPLES);

	for (unsigned i = 0; i < N; ++i) {
		sum_x += s.samples[i];
	}

	double dc = sum_x / N;
	for (unsigned i = 0; i < N; ++i) {
		s.samples[i] -= dc;
	}

	fprintf(stderr, "-dc(%.2f)... ", dc);
}

void normalize_gain(Sample &s)
{
	double sum_x2 = 0.0;
	size_t N = std::min(s.samples.size(), (size_t)FEATURE_SAMPLES);

	for (unsigned i = 0; i < N; ++i) {
		sum_x2 += s.samples[i] * s.samples[i];
	}

	double gain = 8192.0 / sqrt(sum_x2 / N);
	
	for (unsigned i = 0; i < N; ++i) {
		s.samples[i] *= gain;
	}

	fprintf(stderr, "gain(%.2fdB)... ", 10.0 * log10(gain));
}

void find_len(Sample &s)
{
	std::vector<double> d;
	d.push_back(s.length);
	s.descriptors["length"] = d;
	
	fprintf(stderr, "len(%.3fs)... ", s.length);
}

// avg / RMS
void find_msratio(Sample &s)
{
	std::vector<double> d;
	double sum_x2 = 0.0, sum_x = 0.0;
	size_t N = std::min(s.samples.size(), (size_t)FEATURE_SAMPLES);

	for (unsigned i = 0; i < N; ++i) {
		double x = s.samples[i];
		sum_x2 += x * x;
		sum_x  += fabs(x);
	}

	double rms = sqrt(sum_x2 / N);
	double avg = sum_x / N;

	d.push_back(avg / rms);
	s.descriptors["msratio"] = d;
}

// number of zero-crossings
void find_zerocrossings(Sample &s)
{
	std::vector<double> d;
	unsigned zc = 0;
	double last_val = 0.0;
	size_t N = std::min(s.samples.size(), (size_t)FEATURE_SAMPLES);

	for (unsigned i = 0; i < N; ++i) {
		double x = s.samples[i];
		if (x > 0.0 && last_val < 0.0)
			++zc;
		else if (x < 0.0 && last_val > 0.0)
			++zc;
		last_val = x;
	}

	d.push_back(zc);
	s.descriptors["zerocrossings"] = d;
}

// number of zero-crossings, with guard band (schmitt triggering)
void find_zerocrossings_guard(Sample &s)
{
	std::vector<double> d;
	unsigned zc = 0;
	double last_val = 0.0;
	size_t N = std::min(s.samples.size(), (size_t)FEATURE_SAMPLES);

	for (unsigned i = 0; i < N; ++i) {
		double x = s.samples[i];
		if (fabs(x) < 2048.0)
			continue;
		if (x > 0.0 && last_val < 0.0)
			++zc;
		else if (x < 0.0 && last_val > 0.0)
			++zc;
		last_val = x;
	}

	d.push_back(zc);
	s.descriptors["zerocrossings_guard"] = d;
}

// rate of signal change
void find_steepness(Sample &s)
{
	std::vector<double> d;
	double sum_delta_x = 0.0;
	size_t N = std::min(s.samples.size(), (size_t)FEATURE_SAMPLES);

	double last_x = s.samples[0];
	for (unsigned i = 1; i < N; ++i) {
		double x = s.samples[i];
		sum_delta_x += fabs(x - last_x);
		last_x = x;
	}

	d.push_back(sum_delta_x / (N-1));
	s.descriptors["steepness"] = d;
}

int main(int argc, char **argv)
{
	FILE *wisdom = fopen("fft-wisdom", "rb");
	if (wisdom == NULL) {
		fprintf(stderr, "Optimizing FFTs for your system; this will take a few minutes\n");
		fprintf(stderr, "even on a reasonably fast machine, so please wait. This will only\n");
		fprintf(stderr, "need to happen once unless you remove the \"fft-wisdom\" file.\n\n");
		init_mfcc();
		init_vp();
		init_centroid();
		
		wisdom = fopen("fft-wisdom", "wb");
		if (wisdom == NULL) {
			perror("fft-wisdom");
			exit(1);
		}
		fftw_export_wisdom_to_file(wisdom);
		fclose(wisdom);
	} else {
		fftw_import_wisdom_from_file(wisdom);
		fclose(wisdom);
	}

	for (int i = 1; i < argc; ++i) {
		Sample s;

		fprintf(stderr, "%s... ", argv[i]);
		if (strstr(argv[i], ".ogg"))
			read_vorbis_file(s, argv[i]);
		else if (strstr(argv[i], ".flac"))
			read_flac_file(s, argv[i]);
		else if (strstr(argv[i], ".mp3"))
			read_mp3_file(s, argv[i]);
		else
			read_wave_file(s, argv[i]);

		remove_dc_offset(s);
		normalize_gain(s);
		find_len(s);
		
		fprintf(stderr, "msratio... ");
		find_msratio(s);

		fprintf(stderr, "zerocrossings... ");
		find_zerocrossings(s);
		
		fprintf(stderr, "zerocrossings_guard... ");
		find_zerocrossings_guard(s);

		fprintf(stderr, "steepness... ");
		find_steepness(s);
		
		fprintf(stderr, "centroid... ");
		find_centroid(s);

		fprintf(stderr, "mfcc... ");
		find_mfcc(s);

		fprintf(stderr, "mfcc_floor1... ");
		find_mfcc_floor1(s);

		fprintf(stderr, "done.\n");

		printf("filename=%s\n", argv[i]);
		for (std::map<std::string, std::vector<double> >::const_iterator j = s.descriptors.begin(); j != s.descriptors.end(); ++j) {
			printf("%s=", j->first.c_str());
			for (unsigned k = 0; k < j->second.size(); ++k) {
				printf("%f", j->second[k]);
				if (k != j->second.size() - 1) {
					printf(",");
				}
			}
			printf("\n");
		}
		printf("\n");
	}
	
	printf("END\n");
	exit(0);
}
