// f1cc.cpp -- extracts F1CC coefficients from the given sample, and computes
// statistical moments from that.

#include <stdio.h>
#include <math.h>

#include "mfcc.h"
#include "f1cc.h"

// Headers imported from vorbisenc, with some adaption to allow
// the C code to be used from C++.
#define class _class
extern "C" {
	#include "vorbis/codec.h"
	#include "codec_internal.h"
	#include "registry.h"
	#include "psy.h"
	#include "misc.h"
	#include "masking.h"
	#include "scales.h"
}
#undef class
#undef max
#undef min

vorbis_look_psy vp_p;
void _vp_mod_tonemask(vorbis_look_psy *p, float *logfft, float *logmask, float global_specmax, float local_specmax);

void find_mfcc_floor1(Sample &s)
{
	unsigned pos = 0;
	unsigned chunk_len = MFCC_WIN_LEN / MFCC_OVERLAP;
	double buf[MFCC_WIN_LEN];

	// the tone curves seem to go haywire from about 3200 mel (about 11khz), so stop there
	double mel_spacing = 3200.0 / double(MFCC_COEFFICIENTS + 1);

	std::vector<double> coeffs[MFCC_COEFFICIENTS];
	double mu[MFCC_COEFFICIENTS];
	float tone[chunk_len];
	float logfft[chunk_len];

	init_mfcc();
	init_vp();

	// fill the first buffer
	for (unsigned i = 0; i < MFCC_WIN_LEN - chunk_len; ++i, ++pos) {
		buf[i + chunk_len] = s.samples[pos];
	}

	for ( ;; ) {
		if (pos + MFCC_WIN_LEN / MFCC_OVERLAP >= FEATURE_SAMPLES)
			break;

		// move everything back a chunk
		memmove(buf, buf + chunk_len, (MFCC_WIN_LEN - chunk_len) * sizeof(double));

		// read a new chunk
		for (unsigned i = 0; i < chunk_len; ++i, ++pos) {
			buf[i + MFCC_WIN_LEN - chunk_len] = s.samples[pos];
		}

		// window it
		for (unsigned i = 0; i < MFCC_WIN_LEN; ++i) {
			fft_in[i] = buf[i] * window[i] * (1.0/32768.0);
		}

		fftw_execute(fft_plan);

		float scale = 4.0f / chunk_len;
		float scale_dB = todB(&scale) + .345;

		// find amplitude maximum
		float temp = abs(fft_out[0]);
		float ampmax = logfft[0] = scale_dB + todB(&temp) + .345;

		//printf("logfft: %f", logfft[0]);
		for (unsigned i = 1; i < MFCC_WIN_LEN / 2; ++i) {
			temp = abs(fft_out[i]);
			logfft[i] = scale_dB + todB(&temp) + .345;
			//printf(" %f", logfft[i]);

			ampmax = std::max(ampmax, temp);
		}
		//printf("\n");

		// tone masking curve
		_vp_mod_tonemask(&vp_p, logfft, tone, ampmax, ampmax);

#if 0
		printf("tone dump: ");
		for (unsigned i = 0; i < chunk_len / 2; ++i) {
			printf("%f ", tone[i]);
		}
		printf("\n");
#endif

#if 0
		// debugging output for gnuplot et al
		if (pos == (MFCC_WIN_LEN / MFCC_OVERLAP) * 860) {
			FILE *f1 = fopen("fft.plot", "w");
			for (unsigned i = 0; i < MFCC_WIN_LEN / 2; ++i) {
				fprintf(f1, "%f %f\n", hz_to_mel(bin_to_hz(i)), logfft[i]);
			}
			fclose(f1);

			FILE *f2 = fopen("tone.plot", "w");
			for (unsigned i = 0; i < MFCC_WIN_LEN / 2; ++i) {
				fprintf(f2, "%f %f\n", hz_to_mel(bin_to_hz(i)), tone[i]);
			}
			fclose(f2);
			
			usleep(50000);
		}
#endif

		// apply the triangular windows
		for (unsigned i = 0; i < MFCC_COEFFICIENTS; ++i) {
			double start = mel_to_hz(mel_spacing * (i    ));
			double mid   = mel_to_hz(mel_spacing * (i + 1));
			double end   = mel_to_hz(mel_spacing * (i + 2));

			double s = 0.0;

			for (int j = int(ceil(hz_to_bin(start))); j <= floor(hz_to_bin(end)); ++j) {
				double f = bin_to_hz(j);
				double energy = fromdB(tone[j]) * fromdB(tone[j]);

				if (f < mid) {
					s += energy * (f - start) / (mid - start);
				} else {
					s += energy * (end - f) / (end - mid);
				}
			}

			mfcc_in[i] = log(s);
		}

		fftw_execute(dct_plan);

		for (unsigned i = 0; i < MFCC_COEFFICIENTS; ++i) {
			coeffs[i].push_back(mfcc_out[i]);
		}
	}
	
	// first find the first zero moment (average)
	{
		std::vector<double> d;
		for (unsigned i = 0; i < MFCC_COEFFICIENTS; ++i) {
			mu[i] = 0.0;
			for (unsigned j = 0; j < coeffs[i].size(); ++j) {
				mu[i] += coeffs[i][j];
			}
			mu[i] /= coeffs[i].size();

			d.push_back(mu[i]);
		}
		s.descriptors["mfcc_f1_avg"] = d;
	}

	// then the other moments (which are central moments)
	for (unsigned m = 2; m <= MFCC_MOMENTS; ++m) {
		std::vector<double> d;

		for (unsigned i = 0; i < MFCC_COEFFICIENTS; ++i) {
			double v = 0.0;
			for (unsigned j = 0; j < coeffs[i].size(); ++j) {
				v += pow(coeffs[i][j] - mu[i], m);
			}
			d.push_back(v / coeffs[i].size());
		}

		char name[32];
		sprintf(name, "mfcc_f1_moment%u", m);
		s.descriptors[name] = d;
	}
	
	// deltify
	{
		std::vector<double> d;
		for (unsigned i = 0; i < MFCC_COEFFICIENTS; ++i) {
			mu[i] = 0.0;
			for (unsigned j = 1; j < coeffs[i].size(); ++j) {
				mu[i] += fabs(coeffs[i][j] - coeffs[i][j - 1]);
			}
			mu[i] /= (coeffs[i].size() - 1);

			d.push_back(mu[i]);
		}
		s.descriptors["mfcc_f1_delta_avg"] = d;
	}
	for (unsigned m = 2; m <= MFCC_MOMENTS; ++m) {
		std::vector<double> d;

		for (unsigned i = 0; i < MFCC_COEFFICIENTS; ++i) {
			double v = 0.0;
			for (unsigned j = 1; j < coeffs[i].size(); ++j) {
				v += pow(fabs(coeffs[i][j] - coeffs[i][j - 1]) - mu[i], m);
			}
			d.push_back(v / (coeffs[i].size() - 1));
		}

		char name[32];
		sprintf(name, "mfcc_f1_delta_moment%u", m);
		s.descriptors[name] = d;
	}
}

// Everything below this line is copied/adapted from libvorbisenc, mostly
// from lib/floor1.c, and is Copyright 1994-2002 by the Xiphophorus Company
// (http://www.xiph.org/). See the COPYING file included with libvorbis for
// the full license.

typedef struct {
	int att[P_NOISECURVES];
	float boost;
	float decay;
} att3;

typedef struct {
	int   pre[PACKETBLOBS];
	int   post[PACKETBLOBS];
	float kHz[PACKETBLOBS];
	float lowpasskHz[PACKETBLOBS];
} adj_stereo;

typedef struct {
	int lo;
	int hi;
	int fixed;
} noiseguard;

typedef struct vp_adjblock{
	int block[P_BANDS];
} vp_adjblock;
typedef struct {
	int data[NOISE_COMPAND_LEVELS];
} compandblock;
typedef struct {
	int data[P_NOISECURVES][17];
} noise3;
#include "modes/psych_44.h"

bool vp_initialized = false;
vorbis_info_psy vp_vi = _psy_info_template;
vorbis_info_psy_global vp_gi = _psy_global_44[0];

void init_vp()
{
	if (vp_initialized)
		return;

	_vp_psy_init(&vp_p, &vp_vi, &vp_gi, MFCC_FLOOR0_DATA_LEN / 2, 44100);

	vp_initialized = true;
}

#define NEGINF -9999.f

/* octave/(8*eighth_octave_lines) x scale and dB y scale */
static void seed_curve(float *seed,
                       const float **curves,
                       float amp,
                       int oc, int n,
                       int linesper,float dBoffset){
  int i,post1;
  int seedptr;
  const float *posts,*curve;

  int choice=(int)((amp+dBoffset-P_LEVEL_0)*.1f);
  choice=std::max(choice,0);
  choice=std::min(choice,P_LEVELS-1);
  posts=curves[choice];
  curve=posts+2;
  post1=(int)posts[1];
  seedptr=oc+(posts[0]-EHMER_OFFSET)*linesper-(linesper>>1);

  for(i=posts[0];i<post1;i++){
    if(seedptr>0){
      float lin=amp+curve[i];
      if(seed[seedptr]<lin)seed[seedptr]=lin;
    }
    seedptr+=linesper;
    if(seedptr>=n)break;
  }
}

static void seed_loop(vorbis_look_psy *p,
                      const float ***curves,
                      const float *f,
                      const float *flr,
                      float *seed,
                      float specmax){
  vorbis_info_psy *vi=p->vi;
  long n=p->n,i;
  float dBoffset=vi->max_curve_dB-specmax;

  /* prime the working vector with peak values */

  for(i=0;i<n;i++){
    float max=f[i];
    long oc=p->octave[i];
    while(i+1<n && p->octave[i+1]==oc){
      i++;
      if(f[i]>max)max=f[i];
    }

    if(max+6.f>flr[i]){
      oc=oc>>p->shiftoc;
      if(oc>=P_BANDS)oc=P_BANDS-1;
      if(oc<0)oc=0;

      seed_curve(seed,
                 curves[oc],
                 max,
                 p->octave[i]-p->firstoc,
                 p->total_octave_lines,
                 p->eighth_octave_lines,
                 dBoffset);
    }
  }
}

static void seed_chase(float *seeds, int linesper, long n){
  long  posstack[n];
  float ampstack[n];
  long   stack=0;
  long   pos=0;
  long   i;

  for(i=0;i<n;i++){
    if(stack<2){
      posstack[stack]=i;
      ampstack[stack++]=seeds[i];
    }else{
      while(1){
        if(seeds[i]<ampstack[stack-1]){
          posstack[stack]=i;
          ampstack[stack++]=seeds[i];
          break;
        }else{
          if(i<posstack[stack-1]+linesper){
            if(stack>1 && ampstack[stack-1]<=ampstack[stack-2] &&
               i<posstack[stack-2]+linesper){
              /* we completely overlap, making stack-1 irrelevant.  pop it */
              stack--;
              continue;
            }
          }
          posstack[stack]=i;
          ampstack[stack++]=seeds[i];
          break;

        }
      }
    }
  }

  /* the stack now contains only the positions that are relevant. Scan
     'em straight through */

  for(i=0;i<stack;i++){
    long endpos;
    if(i<stack-1 && ampstack[i+1]>ampstack[i]){
      endpos=posstack[i+1];
    }else{
      endpos=posstack[i]+linesper+1; /* +1 is important, else bin 0 is
                                        discarded in short frames */
    }
    if(endpos>n)endpos=n;
    for(;pos<endpos;pos++)
      seeds[pos]=ampstack[i];
  }

  /* there.  Linear time.  I now remember this was on a problem set I
     had in Grad Skool... I didn't solve it at the time ;-) */

}

/* bleaugh, this is more complicated than it needs to be */
#include<stdio.h>
static void max_seeds(vorbis_look_psy *p,
                      float *seed,
                      float *flr){
  long   n=p->total_octave_lines;
  int    linesper=p->eighth_octave_lines;
  long   linpos=0;
  long   pos;

  seed_chase(seed,linesper,n); /* for masking */

  pos=p->octave[0]-p->firstoc-(linesper>>1);

  while(linpos+1<p->n){
    float minV=seed[pos];
    long end=((p->octave[linpos]+p->octave[linpos+1])>>1)-p->firstoc;
    if(minV>p->vi->tone_abs_limit)minV=p->vi->tone_abs_limit;
    while(pos+1<=end){
      pos++;
      if((seed[pos]>NEGINF && seed[pos]<minV) || minV==NEGINF)
        minV=seed[pos];
    }

    end=pos+p->firstoc;
    for(;linpos<p->n && p->octave[linpos]<=end;linpos++)
      if(flr[linpos]<minV)flr[linpos]=minV;
  }

  {
    float minV=seed[p->total_octave_lines-1];
    for(;linpos<p->n;linpos++)
      if(flr[linpos]<minV)flr[linpos]=minV;
  }

}

void _vp_mod_tonemask(vorbis_look_psy *p, float *logfft, float *logmask, float global_specmax, float local_specmax)
{
	int i, n = p->n;
	float seed[p->total_octave_lines];

	for (i = 0; i < p->total_octave_lines; i++)
		seed[i] = NEGINF;
#if 1
	for (i = 0; i < n; i++)
		logmask[i] = local_specmax + p->vi->ath_adjatt;
#else
	float att = local_specmax + p->vi->ath_adjatt;
	if (att < p->vi->ath_maxatt)
		att = p->vi->ath_maxatt;
	for (i = 0; i < n; i++)
		logmask[i] = p->ath[i] + att;
#endif

	/* tone masking */
	seed_loop(p, (const float ***)p->tonecurves, logfft, logmask, seed, global_specmax);
	max_seeds(p, seed, logmask);
}

// End of libvorbisenc code.
