// flac.cpp -- reads in FLAC files using libflac (1.1.4 or newer).

#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <FLAC/all.h>

#include "flac.h"

extern "C" FLAC__StreamDecoderWriteStatus flac_write_callback(const FLAC__StreamDecoder *decoder, const FLAC__Frame *frame, const FLAC__int32 *const buffer[], void *client_data)
{
	Sample *s = (Sample *)client_data;
	assert(frame->header.channels == 2);
	assert(frame->header.bits_per_sample == 16);
	assert(frame->header.sample_rate == 44100);
	
	const FLAC__int32 *lptr = buffer[0], *rptr = buffer[1];

	for (unsigned i = 0; i < frame->header.blocksize; ++i) {
		s->samples.push_back(0.5 * (double(*lptr++) + double(*rptr++)));
	}

	return FLAC__STREAM_DECODER_WRITE_STATUS_CONTINUE;
}

extern "C" void flac_metadata_callback(const FLAC__StreamDecoder *decoder, const FLAC__StreamMetadata *metadata, void *client_data)
{
	if (metadata->type != FLAC__METADATA_TYPE_STREAMINFO)
		return;

	Sample *s = (Sample *)client_data;
	s->length = metadata->data.stream_info.total_samples / 44100.0;
}

extern "C" void flac_error_callback(const FLAC__StreamDecoder *decoder, FLAC__StreamDecoderErrorStatus status, void *client_data)
{
	fprintf(stderr, "FLAC status %d, exiting\n", status);
	exit(1);
}

void read_flac_file(Sample &s, char *filename)
{
	s.samples.reserve(5000000);
	s.length = -1.0;

	FILE *fle = fopen(filename, "rb");
	if (fle == NULL) {
		perror(filename);
		exit(1);
	}

	FLAC__StreamDecoder *dec = FLAC__stream_decoder_new();
	if (FLAC__stream_decoder_init_FILE(dec, fle, flac_write_callback, flac_metadata_callback, flac_error_callback, &s) != FLAC__STREAM_DECODER_INIT_STATUS_OK) {
		fprintf(stderr, "FLAC__stream_decoder_init_FILE failed.\n");
		exit(1);
	}

	/*
	 * Decode until we have enough samples _and_ we know the length. Hopefully
	 * the length will come from metadata; if not, we will decode the entire
	 * stream to memory. This could be avoided, but it's not a use-case we need
	 * to worry particularly about in this context.
	 */
	while (s.samples.size() < FEATURE_SAMPLES || s.length < 0.0) {
		if (FLAC__stream_decoder_get_state(dec) == FLAC__STREAM_DECODER_END_OF_STREAM)
			break;

		if (FLAC__stream_decoder_process_single(dec) == false) {
			fprintf(stderr, "FLAC__stream_decoder_process_until_end_of_stream failed.\n");
			exit(1);
		}
	}
	FLAC__stream_decoder_finish(dec);
	FLAC__stream_decoder_delete(dec);
}
