// mfcc.cpp -- extracts MFCC coefficients from the given sample, and computes
// statistical moments from that.

#include "mfcc.h"

bool mfcc_initialized = false;
fftw_plan fft_plan, dct_plan;
double *fft_in;
std::complex<double> *fft_out;
double *mfcc_in, *mfcc_out;
double *window;

void init_mfcc()
{
	if (mfcc_initialized)
		return;

	// FFT
	fft_in = reinterpret_cast<double *> (fftw_malloc(sizeof(double) * MFCC_WIN_LEN));
	fft_out = reinterpret_cast<std::complex<double> *> (fftw_malloc(sizeof(std::complex<double>) * (MFCC_WIN_LEN / 2 + 1)));
	fft_plan = fftw_plan_dft_r2c_1d(MFCC_WIN_LEN, fft_in, reinterpret_cast<fftw_complex *> (fft_out), FFTW_PATIENT | FFTW_DESTROY_INPUT);

	// DCT
	mfcc_in = reinterpret_cast<double *> (fftw_malloc(sizeof(double) * MFCC_COEFFICIENTS));
	mfcc_out = reinterpret_cast<double *> (fftw_malloc(sizeof(double) * MFCC_COEFFICIENTS));
	dct_plan = fftw_plan_r2r_1d(MFCC_COEFFICIENTS, mfcc_in, mfcc_out, FFTW_REDFT10, FFTW_PATIENT | FFTW_DESTROY_INPUT);

	// Initialize the Hamming window
	window = new double[MFCC_WIN_LEN];
	for (unsigned i = 0; i < MFCC_WIN_LEN; ++i) {
		window[i] = 0.54 - 0.46 * cos(2.0 * M_PI * double(i) / double(MFCC_WIN_LEN - 1));
	}

	mfcc_initialized = true;
}

void find_mfcc(Sample &s)
{
	unsigned pos = 0;
	unsigned chunk_len = MFCC_WIN_LEN / MFCC_OVERLAP;
	double buf[MFCC_WIN_LEN];
	double mel_spacing = hz_to_mel(44100.0 / 2.0) / double(MFCC_COEFFICIENTS + 1);  // ??
	std::vector<double> coeffs[MFCC_COEFFICIENTS];
	double mu[MFCC_COEFFICIENTS];

	init_mfcc();

	// fill the first buffer
	for (unsigned i = 0; i < MFCC_WIN_LEN - chunk_len; ++i, ++pos) {
		buf[i + chunk_len] = s.samples[pos];
	}

	for ( ;; ) {
		if (pos + MFCC_WIN_LEN / MFCC_OVERLAP >= FEATURE_SAMPLES)
			break;

		// move everything back a chunk
		memmove(buf, buf + chunk_len, (MFCC_WIN_LEN - chunk_len) * sizeof(double));

		// read a new chunk
		for (unsigned i = 0; i < chunk_len; ++i, ++pos) {
			buf[i + MFCC_WIN_LEN - chunk_len] = s.samples[pos];
		}

		// window it
		for (unsigned i = 0; i < MFCC_WIN_LEN; ++i) {
			fft_in[i] = buf[i] * window[i];
		}

		fftw_execute(fft_plan);

		// apply the triangular windows
		for (unsigned i = 0; i < MFCC_COEFFICIENTS; ++i) {
			double start = mel_to_hz(mel_spacing * (i    ));
			double mid   = mel_to_hz(mel_spacing * (i + 1));
			double end   = mel_to_hz(mel_spacing * (i + 2));

			double s = 0.0;

			for (int j = int(ceil(hz_to_bin(start))); j <= floor(hz_to_bin(end)); ++j) {
				double f = bin_to_hz(j);
				double energy = abs(fft_out[j]) * abs(fft_out[j]);

				if (f < mid) {
					s += energy * (f - start) / (mid - start);
				} else {
					s += energy * (end - f) / (end - mid);
				}
			}

			mfcc_in[i] = log(s + 0.001);
		}

		fftw_execute(dct_plan);

		for (unsigned i = 0; i < MFCC_COEFFICIENTS; ++i) {
			coeffs[i].push_back(mfcc_out[i]);
		}
	}

#if 0
	// output a little grayscale image
	{
		FILE *pgm = fopen("out.pgm", "wb");
		fprintf(pgm, "P5\n%u %u\n255\n", coeffs[0].size() - 1, MFCC_COEFFICIENTS);
				
		double max = 0.0;
		for (unsigned i = 0; i < MFCC_COEFFICIENTS; ++i) {
			for (unsigned j = 0; j < coeffs[0].size(); ++j) {
				max = std::max(max, log(fabs(coeffs[i][j])));
			}
		}
		for (unsigned i = 0; i < MFCC_COEFFICIENTS; ++i) {
			for (unsigned j = 1; j < coeffs[0].size(); ++j) {
				double z = 255.0 * log(fabs(coeffs[i][j])) / max;
				if (z < 0.0)
					fprintf(pgm, "%c", 0);
				else
					fprintf(pgm, "%c", (unsigned char)z);
			}
		}

		fclose(pgm);
	}

	// output a graph of the 7th MFCC coeffient
	{
		draw_graph("graph.pgm", coeffs[6]);
		
		std::vector<double> c_level1;
		std::vector<double> d_level1;
		
		{
			std::vector<double> tmp;
			wavelet_filter_haar(coeffs[6], tmp);
			decimate(tmp, c_level1, 2);
		}
		{
			std::vector<double> tmp;
			wavelet_filter_haar_detail(coeffs[6], tmp);
			decimate(tmp, d_level1, 2);
		}
		
		std::vector<double> c_level2;
		std::vector<double> d_level2;
		
		{
			std::vector<double> tmp;
			wavelet_filter_haar(c_level1, tmp);
			decimate(tmp, c_level2, 2);
		}
		{
			std::vector<double> tmp;
			wavelet_filter_haar_detail(c_level1, tmp);
			decimate(tmp, d_level2, 2);
		}

		draw_graph("c_level2.pgm", c_level2);
		draw_graph("d_level2.pgm", d_level2);
		
		std::vector<double> c_level3;
		std::vector<double> d_level3;
		
		{
			std::vector<double> tmp;
			wavelet_filter_haar(c_level2, tmp);
			decimate(tmp, c_level3, 2);
		}
		{
			std::vector<double> tmp;
			wavelet_filter_haar_detail(c_level2, tmp);
			decimate(tmp, d_level3, 2);
		}

		draw_graph("c_level3.pgm", c_level3);
		draw_graph("d_level3.pgm", d_level3);
		
		std::vector<double> c_level4;
		std::vector<double> d_level4;
		
		{
			std::vector<double> tmp;
			wavelet_filter_haar(c_level3, tmp);
			decimate(tmp, c_level4, 2);
		}
		{
			std::vector<double> tmp;
			wavelet_filter_haar_detail(c_level3, tmp);
			decimate(tmp, d_level4, 2);
		}

		draw_graph("c_level4.pgm", c_level4);
		draw_graph("d_level4.pgm", d_level4);
		
		std::vector<double> c_level5;
		std::vector<double> d_level5;
		
		{
			std::vector<double> tmp;
			wavelet_filter_haar(c_level4, tmp);
			decimate(tmp, c_level5, 2);
		}
		{
			std::vector<double> tmp;
			wavelet_filter_haar_detail(c_level4, tmp);
			decimate(tmp, d_level5, 2);
		}

		draw_graph("c_level5.pgm", c_level5);
		draw_graph("d_level5.pgm", d_level5);
		
		std::vector<double> c_level6;
		std::vector<double> d_level6;
		
		{
			std::vector<double> tmp;
			wavelet_filter_haar(c_level5, tmp);
			decimate(tmp, c_level6, 2);
		}
		{
			std::vector<double> tmp;
			wavelet_filter_haar_detail(c_level5, tmp);
			decimate(tmp, d_level6, 2);
		}

		draw_graph("c_level6.pgm", c_level6);
		draw_graph("d_level6.pgm", d_level6);
	}
#endif

	// first find the first zero moment (average)
	{
		std::vector<double> d;
		for (unsigned i = 0; i < MFCC_COEFFICIENTS; ++i) {
			mu[i] = 0.0;
			for (unsigned j = 0; j < coeffs[i].size(); ++j) {
				mu[i] += coeffs[i][j];
			}
			mu[i] /= coeffs[i].size();

			d.push_back(mu[i]);
		}
		s.descriptors["mfcc_avg"] = d;
	}

	// then the other moments (which are central moments)
	for (unsigned m = 2; m <= MFCC_MOMENTS; ++m) {
		std::vector<double> d;

		for (unsigned i = 0; i < MFCC_COEFFICIENTS; ++i) {
			double v = 0.0;
			for (unsigned j = 0; j < coeffs[i].size(); ++j) {
				v += pow(coeffs[i][j] - mu[i], m);
			}
			d.push_back(v / coeffs[i].size());
		}

		char name[32];
		sprintf(name, "mfcc_moment%u", m);
		s.descriptors[name] = d;
	}

	// deltify
	{
		std::vector<double> d;
		for (unsigned i = 0; i < MFCC_COEFFICIENTS; ++i) {
			mu[i] = 0.0;
			for (unsigned j = 1; j < coeffs[i].size(); ++j) {
				mu[i] += fabs(coeffs[i][j] - coeffs[i][j - 1]);
			}
			mu[i] /= (coeffs[i].size() - 1);

			d.push_back(mu[i]);
		}
		s.descriptors["mfcc_delta_avg"] = d;
	}
	for (unsigned m = 2; m <= MFCC_MOMENTS; ++m) {
		std::vector<double> d;

		for (unsigned i = 0; i < MFCC_COEFFICIENTS; ++i) {
			double v = 0.0;
			for (unsigned j = 1; j < coeffs[i].size(); ++j) {
				v += pow(fabs((coeffs[i][j] - coeffs[i][j - 1])) - mu[i], m);
			}
			d.push_back(v / (coeffs[i].size() - 1));
		}

		char name[32];
		sprintf(name, "mfcc_delta_moment%u", m);
		s.descriptors[name] = d;
	}
}

