/*
 * Simple MP3 decoder using MAD. We use the low-level API (documented in
 * madlld) so we can stop synthesizing PCM and only count time when we have
 * the data we need; a bit more involved, but much faster. Heavily based
 * on madlld.c.
 */

#include <stdio.h>
#include <assert.h>
#include <mad.h>

#include "mp3.h"

static inline
signed int scale(mad_fixed_t sample)
{
	/* round */
	sample += (1L << (MAD_F_FRACBITS - 16));

	/* clip */
	if (sample >= MAD_F_ONE)
		sample = MAD_F_ONE - 1;
	else if (sample < -MAD_F_ONE)
		sample = -MAD_F_ONE;

	/* quantize */
	return sample >> (MAD_F_FRACBITS + 1 - 16);
}

// Read in the entire file, which is the simplest way to provide
// the EOF semantics libmad assumes.
char *gobble_file(char *filename, long *file_len)
{
	FILE *fle = fopen(filename, "rb");
	if (fle == NULL) {	
		perror(filename);
		exit(1);
	}

	fseek(fle, 0, SEEK_END);
	*file_len = ftell(fle);

	char *mp3_buf = new char[*file_len + 8]; // eight guard bytes
	rewind(fle);
	fread(mp3_buf, *file_len, 1, fle);
	fclose(fle);

	memset(mp3_buf + *file_len, 0, 8);

	return mp3_buf;
}

void read_mp3_file(Sample &s, char *filename)
{
	s.samples.reserve(FEATURE_SAMPLES + 44100); // some extra room

	unsigned mp3_pos = 0;
	long mp3_len;
	char *mp3_buf = gobble_file(filename, &mp3_len);

	struct mad_stream stream;
	struct mad_frame frame;
	struct mad_synth synth;  
	mad_timer_t timer;

	mad_stream_init(&stream);
	mad_frame_init(&frame);
	mad_synth_init(&synth);
	mad_timer_reset(&timer);

	for ( ;; ) {
		// need to fill buffer?
		if (stream.buffer == NULL || stream.error == MAD_ERROR_BUFLEN) {
			if (stream.next_frame != NULL) {
				size_t remaining = stream.bufend - stream.next_frame;
				mp3_pos -= remaining;
			}
	
			mad_stream_buffer(&stream, (unsigned char *)mp3_buf + mp3_pos, mp3_len + 8 - mp3_pos);
			stream.error = MAD_ERROR_NONE;
		}

		if (mad_frame_decode(&frame, &stream)) {
			if (MAD_RECOVERABLE(stream.error) && stream.error == MAD_ERROR_LOSTSYNC && stream.this_frame == (unsigned char *)mp3_buf + mp3_len) {
				// end of stream
				break;
			} else {
				fprintf(stderr, "mad error. exiting.\n");
				exit(1);
			}
		}

		mad_timer_add(&timer, frame.header.duration);

		if (timer.seconds < 30) {
			mad_synth_frame(&synth, &frame);
			assert(MAD_NCHANNELS(&frame.header)==2);
			
			for (unsigned i = 0; i < synth.pcm.length; ++i) {
				signed short l, r;
			
				l = scale(synth.pcm.samples[0][i]);
				r = scale(synth.pcm.samples[1][i]);

#if 0
				/*
				 * There's a lot of codec delay in MP3. This skips
				 * the first 1105 samples (LAME's estimate of the
				 * delay for our test MP3s) to make lining up plots
				 * easier.
				 */
				static unsigned delayed = 0;
				if (++delayed < 1105)
					continue;
#endif

				s.samples.push_back(0.5 * (double(l) + double(r)));
			}
		}
	}

	delete[] mp3_buf;

	s.length = double(timer.seconds) + double(timer.fraction) / double(MAD_TIMER_RESOLUTION);
}
