// Simple descriptor tester -- tests a given descriptor configuration
// using the nearest-neighbor test described in the report.

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <map>
#include <string>
#include <vector>
#include <algorithm>

// Basic data structures.
typedef std::map<std::string, std::vector<double> > descriptor;
struct song {
	descriptor desc;
	unsigned file_num, subfile_num;
	std::string filename;
	unsigned class_id;
};

// The collection of all song data, unprocessed.
std::vector<song> songs;

// All vectors for references (r) and others (s). The "trans" versions
// are transformed by the covariance matrix (inv_cov_mat, below), if any.
// rnum and snum are counts.
double *rvec, *svec, *rtransvec, *stransvec;
int *rnum, *snum;

// How many references are there in all? (Non-references is songs.size()
// minus this.)
unsigned num_ref = 0;

// Various filenames and options.
static std::string output_data_filename;
static std::string output_class_filename;
static std::string output_tex_filename;
static std::string output_raw_filename;
static std::string input_matrix_filename;
static std::string input_center_filename;
static std::string input_scale_filename;
static std::string input_inv_cov_filename;
static bool show_misclassified = false;
static std::string dump_filename;
static bool do_read_dump = false, do_write_dump = false;
static int num_dimensions = -1;

// Various matrices, usually read from tet files.
double *rot_mat = NULL, *center = NULL, *scale = NULL, *inv_cov_mat = NULL;

// Options to be given to getopt_long(). 
const static struct option longopts[] = {
	{ "features", required_argument, NULL, 'f' },
	{ "output-data", required_argument, NULL, 'o' },
	{ "output-class", required_argument, NULL, 'O' },
	{ "output-tex", required_argument, NULL, 'x' },
	{ "output-raw", required_argument, NULL, 'X' },
	{ "show-misclassified", no_argument, NULL, 'm' },
	{ "write-dump", required_argument, NULL, 'w' },
	{ "read-dump", required_argument, NULL, 'r' },
	{ "num-dimensions", required_argument, NULL, 'd' },
	{ "transform-matrix", required_argument, NULL, 't' },
	{ "transform-center", required_argument, NULL, 'c' },
	{ "transform-scale", required_argument, NULL, 's' },
	{ "inv-covariance-matrix", required_argument, NULL, 'i' },
};

// Read a descriptor file as generated by the feature extractor.
void read_file(const char * const filename)
{
	static unsigned file_num = 0;
	unsigned subfile_num = 0;
	std::string song_filename;

	descriptor d;
	FILE *f = fopen(filename, "r");
	if (f == NULL) {
		perror(filename);
		exit(1);
	}

	for ( ;; ) {
		char buf[4096];
		char *ptr = fgets(buf, 4096, f);
		if (ptr == NULL) {
			fprintf(stderr, "%s: Unexpected EOF\n", filename);
			exit(1);
		}

		// strip trailing EOL
		ptr = strchr(buf, '\n');
		if (ptr != NULL)
			*ptr = '\0';

		// EOF marker
		if (strcmp(buf, "END") == 0)
			break;

		// end-of-record marker
		if (strlen(buf) == 0) {
			song s;
			s.file_num = file_num;
			s.subfile_num = subfile_num++;
			s.desc = d;
			s.filename = song_filename;

			// Hard-coded class list
			if (strstr(song_filename.c_str(), ".flac") != NULL) {
				s.class_id = 0;
			} else if (strstr(song_filename.c_str(), ".128kbps.mp3") != NULL) {
				s.class_id = 1;
			} else if (strstr(song_filename.c_str(), ".192kbps.mp3") != NULL) {
				s.class_id = 2;
			} else if (strstr(song_filename.c_str(), ".256kbps.mp3") != NULL) {
				s.class_id = 3;
			} else if (strstr(song_filename.c_str(), ".64kbps.ogg") != NULL) {
				s.class_id = 4;
			} else if (strstr(song_filename.c_str(), ".128kbps.ogg") != NULL) {
				s.class_id = 5;
			} else if (strstr(song_filename.c_str(), ".192kbps.ogg") != NULL) {
				s.class_id = 6;
			} else {
				fprintf(stderr, "Unable to classify file '%s'\n", 
					song_filename.c_str());
				exit(1);
			}

			songs.push_back(s);
			d = descriptor();
			continue;
		}

		ptr = strchr(buf, '=');
		if (ptr == NULL) {
			fprintf(stderr, "%s: Malformed line '%s'\n", filename, buf);
			exit(1);
		}

		*ptr = '\0';
		char *key = buf;
		char *value = ptr + 1;

		if (strcmp(key, "filename") == 0) {
			song_filename = value;
			continue;
		}

		std::vector<double> values;
		values.reserve(16);

		ptr = strtok(value, ",");
		while (ptr != NULL) {
			double v = atof(ptr);

			// some fiddling that should really be done in descriptor.cpp
			if (strcmp(key, "mfcc_moment2") == 0 || strcmp(key, "mfcc_f1_moment2") == 0)
				v = sqrt(v);
			else if (strcmp(key, "mfcc_moment3") == 0 || strcmp(key, "mfcc_f1_moment3") == 0)
				v = cbrt(v);
			else if (strcmp(key, "mfcc_moment4") == 0 || strcmp(key, "mfcc_f1_moment4") == 0)
				v = pow(v, 1.0/4.0);
			else if (strcmp(key, "mfcc_moment5") == 0 || strcmp(key, "mfcc_f1_moment5") == 0)
				v = (v < 0.0) ? -pow(-v, 1.0/5.0) : pow(v, 1.0/5.0);

			values.push_back(v);
			ptr = strtok(NULL, ",");
		}

		d.insert(std::make_pair(key, values));
	}

	fclose(f);
	++file_num;
	
	if (file_num % 64 == 0)
		fprintf(stderr, "Reading: %u\r", songs.size());
}
			
// Read several files -- the file names are first read from a file list,
// and each file name is then passed to read_file().
void read_filelist(char *filename)
{
	FILE *f = fopen(filename, "r");
	if (f == NULL) {
		perror(filename);
		exit(1);
	}

	for ( ;; ) {
		char buf[4096];
		char *ptr = fgets(buf, 4096, f);
		if (ptr == NULL)
			break;
		
		// strip trailing EOL
		ptr = strchr(buf, '\n');
		if (ptr != NULL)
			*ptr = '\0';

		read_file(buf);
	}

	fclose(f);
}

// Extract a single feature from a "song" struct. A feature can either
// be of the form "name[index]", or just "name" (in which case index
// is taken to be zero). No error checking is performed, in particular
// not on out-of-bounds indices.
double extract_feature(song &s, std::string feature)
{
	char buf[256];
	unsigned i = feature.find_first_of("[");
	if (i == std::string::npos) {
		return s.desc[feature][0];
	} else {
		strncpy(buf, feature.c_str(), i);
		buf[i] = '\0';

		return s.desc[buf][atoi(feature.c_str() + i + 1)];
	}
}

// Euclidean distance between two vectors of dimensionality LEN.
double distance(double *p1, double *p2, unsigned len)
{
	double sum = 0.0;
	for (unsigned i = 0; i < len; ++i, ++p1, ++p2) {
		sum += (*p2 - *p1) * (*p2 - *p1);
	}

	return sum;
}

// Mahalanobis distance between two vectors of dimensionality LEN. p1t and p2t
// are taken to be transformed (by the inverse covariance matrix) versions of p1
// and p2, respectively.
double distancem(double *p1, double *p2, double *p1t, double *p2t, unsigned len)
{
	double sum = 0.0;
	for (unsigned i = 0; i < len; ++i, ++p1, ++p2, ++p1t, ++p2t) {
		sum += (*p2 - *p1) * (*p2t - *p1t);
	}

	return sum;
}

// Split a feature string (feature[1],other_feature[0],etc.) into a list of single
// feature names.
std::vector<std::string> parse_features(const char * const feature_string)
{
	std::vector<std::string> features;
	char *fs = strdup(feature_string);

	char *ptr = strtok(fs, ",");
	while (ptr != NULL) {
		features.push_back(ptr);
		ptr = strtok(NULL, ",");
	}

	free(fs);
	return features;
}

// From the list of songs and the list of desired features, create a simple
// two-dimensional arrays of just the features we need for processing. This
// structure is much smaller, simpler and faster to work on than the general-
// purpose "song" structure.
void extract_features(const std::vector<std::string> &features)
{
	double tmp[features.size()];

	// count the number of references
	for (unsigned i = 0; i < songs.size(); ++i) {
		if (songs[i].class_id == 0)
			++num_ref;
	}

	// extract the feature vectors
	rvec = new double[num_ref * features.size()];
	svec = new double[(songs.size() - num_ref) * features.size()];
	rnum = new int[num_ref];
	snum = new int[songs.size() - num_ref];

	FILE *fp = NULL, *cl = NULL;

	if (output_data_filename.size() > 0) {
		fp = fopen(output_data_filename.c_str(), "w");
		if (fp == NULL) {
			perror(output_data_filename.c_str());
			exit(1);
		}
		for (unsigned j = 0; j < features.size(); ++j) {
			fprintf(fp, "%s", features[j].c_str());
			if (j == features.size() - 1)
				fprintf(fp, "\n");
			else
				fprintf(fp, ",");
		}
	}
	if (output_class_filename.size() > 0) {
		cl = fopen(output_class_filename.c_str(), "w");
		if (cl == NULL) {
			perror(output_class_filename.c_str());
			exit(1);
		}
		fprintf(cl, "class\n");
	}

	for (unsigned i = 0, m = 0, n = 0; i < songs.size(); ++i) {
		double *ptr;
		if (songs[i].class_id == 0) {
			ptr = rvec + m * features.size();
			rnum[m] = i;
			++m;
		} else {
			ptr = svec + n * features.size();
			snum[n] = i;
			++n;
		}

		for (unsigned j = 0; j < features.size(); ++j) {
			ptr[j] = extract_feature(songs[i], features[j]);
		}

		// Apply centering, scaling and rotation if applicable.
		if (center) {
			for (unsigned j = 0; j < features.size(); ++j) {
				ptr[j] -= center[j];
			}
		}
		if (scale) {
			for (unsigned j = 0; j < features.size(); ++j) {
				ptr[j] *= scale[j];
			}
		}
		if (rot_mat) {
			for (unsigned j = 0; j < features.size(); ++j) {
				tmp[j] = 0.0;
				for (unsigned k = 0; k < features.size(); ++k) {
					tmp[j] += rot_mat[j * features.size() + k] * ptr[j];
				}
			}
			memcpy(ptr, tmp, features.size() * sizeof(double));
		}

		if (fp) {
			for (unsigned j = 0; j < features.size(); ++j) {
				fprintf(fp, "%f", ptr[j]);
				if (j == features.size() - 1)
					fprintf(fp, "\n");
				else
					fprintf(fp, ",");
			}
		}
		if (cl) {
			fprintf(cl, "%u\n", songs[i].file_num);
		}
	}

	if (fp)
		fclose(fp);
	if (cl)
		fclose(cl);
}

// Precalculate transformed versions of all vectors, given an inverse covariance
// matrix.
double *precalc_transformed_features(double *vecs, double *matrix, unsigned num_vec, unsigned num_features)
{
	double *ret = new double[num_vec * num_features];

	for (unsigned i = 0; i < num_vec; ++i) {
		for (unsigned j = 0; j < num_features; ++j) {
			double sum = 0.0;
			for (unsigned k = 0; k < num_features; ++k) {
				sum += vecs[i * num_features + k] * matrix[j * num_features + k];
			}
			ret[i * num_features + j] = sum;
		}
	}

	return ret;
}

// Perform the actual classification, matching all non-references against all
// references.
void do_search(unsigned num_features, unsigned stride)
{
	// nearest-neighbor search
	unsigned correct = 0, total = 0;
	unsigned correct_class[7], total_class[7];

	for (unsigned i = 0; i < 7; ++i) {
		correct_class[i] = total_class[i] = 0;
	}

	for (unsigned i = 0; i < (songs.size() - num_ref); ++i) {
		int best_match = -1;
		double best_match_val = HUGE_VAL;

		for (unsigned j = 0; j < num_ref; ++j) {
			double this_match_val;
			
			if (rtransvec == NULL) {
				// Euclidian distance
				this_match_val = distance(svec + i * stride, rvec + j * stride, num_features);
			} else {
				// Mahalanobis distance
				this_match_val = distancem(svec + i * stride, rvec + j * stride, stransvec + i * stride, rtransvec + j * stride, num_features);
			}

			if (best_match == -1 || this_match_val < best_match_val) {
				best_match = j;
				best_match_val = this_match_val;
			}
		}

		if (songs[snum[i]].file_num == songs[rnum[best_match]].file_num) {
			++correct;
			++correct_class[songs[snum[i]].class_id];
		} else {
			if (show_misclassified) 
				fprintf(stderr, "Misclassified: '%s' => '%s'\n", songs[snum[i]].filename.c_str(),
					songs[rnum[best_match]].filename.c_str());
		}
		++total;
		++total_class[songs[snum[i]].class_id];

		if (total % 64 == 0)
			fprintf(stderr, "%5u/%5u (%5.1f%%)... \r", correct, total, (100.0 * correct / total));
	}

	if (output_tex_filename.size() > 0) {
		FILE *f = fopen(output_tex_filename.c_str(), "wb");
		if (f == NULL) {
			perror(output_tex_filename.c_str());
			exit(1);
		}

		for (unsigned i = 1; i <= 6; ++i) {
			fprintf(f, "%6.1f\\%% & ", 100.0 * (1.0 - correct_class[i] / total_class[i]));
		}

		fprintf(f, "%6.1f\\%% \n", 100.0 * (1.0 - correct / total));

		fclose(f);
	}
	if (output_raw_filename.size() > 0) {
		FILE *f = fopen(output_raw_filename.c_str(), "wb");
		if (f == NULL) {
			perror(output_raw_filename.c_str());
			exit(1);
		}

		for (unsigned i = 1; i <= 6; ++i) {
			fprintf(f, "%u %u\n", correct_class[i], total_class[i]);
		}

		fprintf(f, "%u %u\n", correct, total);

		fclose(f);
	}
	printf("Classifier was right %.1f%% of the time (%u/%u).\n", (100.0 * correct / total), correct, total);
}

// Serializes the "songs" array down to a single file for quicker reading back.
void write_dump(const std::string &filename)
{
	FILE *f = fopen(filename.c_str(), "wb");
	if (f == NULL) {
		perror(filename.c_str());
		exit(1);
	}

	unsigned num_songs = songs.size();
	fwrite(&num_songs, sizeof(num_songs), 1, f);

	for (std::vector<song>::const_iterator i = songs.begin(); i != songs.end(); ++i) {
		fwrite(&i->file_num, sizeof(i->file_num), 1, f);
		fwrite(&i->subfile_num, sizeof(i->subfile_num), 1, f);
		fwrite(&i->class_id, sizeof(i->class_id), 1, f);

		unsigned filename_len = i->filename.size();
		fwrite(&filename_len, sizeof(filename_len), 1, f);
		fwrite(i->filename.c_str(), filename_len, 1, f);

		unsigned num_descriptors = i->desc.size();
		fwrite(&num_descriptors, sizeof(num_descriptors), 1, f);

		for (std::map<std::string, std::vector<double> >::const_iterator j = i->desc.begin(); j != i->desc.end(); ++j) {
			unsigned key_len = j->first.size();
			fwrite(&key_len, sizeof(key_len), 1, f);
			fwrite(j->first.c_str(), key_len, 1, f);

			unsigned num_elem = j->second.size();
			fwrite(&num_elem, sizeof(num_elem), 1, f);

			for (std::vector<double>::const_iterator k = j->second.begin(); k != j->second.end(); ++k) {
				fwrite(&*k, sizeof(*k), 1, f);
			}
		}
	}
	fclose(f);
}

// Reads the "songs" array back from a file created by write_dump().
void read_dump(const std::string &filename)
{
	char buf[4096]; // vulnerable to overflows, but we're only reading in trusted data anyhow
	FILE *f = fopen(filename.c_str(), "rb");
	if (f == NULL) {
		perror(filename.c_str());
		exit(1);
	}

	unsigned num_songs;
	fread(&num_songs, sizeof(num_songs), 1, f);

	songs.reserve(num_songs);

	for (unsigned i = 0; i < num_songs; ++i) {
		song s;

		fread(&s.file_num, sizeof(s.file_num), 1, f);
		fread(&s.subfile_num, sizeof(s.subfile_num), 1, f);
		fread(&s.class_id, sizeof(s.class_id), 1, f);

		unsigned filename_len;
		fread(&filename_len, sizeof(filename_len), 1, f);
		fread(buf, filename_len, 1, f);
		buf[filename_len] = '\0';
		s.filename = buf;

		unsigned num_descriptors;
		fread(&num_descriptors, sizeof(num_descriptors), 1, f);

		for (unsigned j = 0; j < num_descriptors; ++j) {
			unsigned key_len;
			fread(&key_len, sizeof(key_len), 1, f);
			fread(buf, key_len, 1, f);
			buf[key_len] = '\0';

			std::string key = buf;

			unsigned num_elem;
			fread(&num_elem, sizeof(num_elem), 1, f);

			std::vector<double> elem(num_elem);

			for (unsigned k = 0; k < num_elem; ++k) {
				double tmp;
				fread(&tmp, sizeof(tmp), 1, f);
				elem[k] = tmp;
			}

			s.desc.insert(std::make_pair(key, elem));
		}

		songs.push_back(s);
	}
	fclose(f);

	fprintf(stderr, "Read %u songs from dump file.\n", songs.size());
}

// Allocate and read a NUM_ELEM-matrix from a text file.
void read_matrix(double **mat, const std::string &filename, unsigned num_elem)
{
	FILE *f = fopen(filename.c_str(), "r");
	if (f == NULL) {
		perror(filename.c_str());
		exit(1);
	}

	*mat = new double[num_elem];

	for (unsigned i = 0; i < num_elem; ++i) {
		double tmp;
		if (fscanf(f, "%lf", &tmp) == 0) {
			fprintf(stderr, "%s contains only %u elements, expected %u\n",
				filename.c_str(), i, num_elem);
			exit(1);
		}
		(*mat)[i] = tmp;
	}

	double tmp;
	if (fscanf(f, "%lf", &tmp) == 0) {
		fprintf(stderr, "%s contains more than the expected %u elements\n",
			filename.c_str(), num_elem);
		exit(1);
	}

	fclose(f);
}

int main(int argc, char **argv)
{
	std::vector<std::string> features;
	
	songs.reserve(70000);

	int option_index = 0;
	for ( ;; ) {
		int c = getopt_long(argc, argv, "f:o:O:x:X:w:r:d:t:c:s:i:m", longopts, &option_index);
		if (c == -1)
			break;

		switch (c) {
		case 'f':
			if (features.size() > 0) {
				fprintf(stderr, "-f given twice; exiting.\n");
				exit(1);
			}
			features = parse_features(optarg);
			break;
		case 'o':
			output_data_filename = optarg;
			break;
		case 'O':
			output_class_filename = optarg;
			break;
		case 'x':
			output_tex_filename = optarg;
			break;
		case 'X':
			output_raw_filename = optarg;
			break;
		case 'w':
			dump_filename = optarg;
			do_write_dump = true;
			break;
		case 'r':
			dump_filename = optarg;
			do_read_dump = true;
			break;
		case 'm':
			show_misclassified = true;
			break;
		case 'd':
			num_dimensions = atoi(optarg);
			break;
		case 't':
			input_matrix_filename = optarg; 
			break;
		case 'c':
			input_center_filename = optarg; 
			break;
		case 's':
			input_scale_filename = optarg; 
			break;
		case 'i':
			input_inv_cov_filename = optarg; 
			break;
		default:
			fprintf(stderr, "Unknown option character '%c'\n", c);
			exit(1);
		}
	}

	if (features.size() == 0 || (optind >= argc && !do_read_dump)) {
		fprintf(stderr, "Usage: voronoi -f FEATURELIST FILES ...");
		exit(1);
	}
	if (do_read_dump && do_write_dump) {
		fprintf(stderr, "Error: --read-dump and --write-dump can not both be specified\n");
		exit(1);
	}
	
	if (do_read_dump) {
		read_dump(dump_filename);
	}
	
	while (optind < argc) {
		if (argv[optind][0] == '@') {
			read_filelist(argv[optind] + 1);
		} else {
			read_file(argv[optind]);
		}
		++optind;
	}

	if (do_write_dump) {
		write_dump(dump_filename);
	}

	if (input_matrix_filename.size() > 0) {
		read_matrix(&rot_mat, input_matrix_filename, features.size() * features.size());
	}
	if (input_center_filename.size() > 0) {
		read_matrix(&center, input_center_filename, features.size());
	}
	if (input_scale_filename.size() > 0) {
		read_matrix(&scale, input_scale_filename, features.size());
		for (unsigned i = 0; i < features.size(); ++i) {
			scale[i] = 1.0 / scale[i];
		}
	}

	extract_features(features);

	if (input_inv_cov_filename.size() > 0) {
		read_matrix(&inv_cov_mat, input_inv_cov_filename, features.size() * features.size());
		rtransvec = precalc_transformed_features(rvec, inv_cov_mat, num_ref, features.size());
		stransvec = precalc_transformed_features(svec, inv_cov_mat, songs.size() - num_ref, features.size());
	}

	if (num_dimensions == -1) {
		do_search(features.size(), features.size());
	} else if (num_dimensions > 0) {
		do_search(std::min((unsigned)num_dimensions, features.size()), features.size());
	} else {	
		fprintf(stderr, "No classification done.\n");
	}
}

