<?php
/**
 * This script combines match scores for several descriptors using the
 * methods outlined in the study. It needs a lot of memory, probably around
 * 512MB or so. 
 * 
 * Inserts the results directly into MySQL.
 * 
 * @author Vegard Andreas Larsen <vegarl@stud.ntnu.no>
 */
require_once('include/include.php');
$oDB = CDatabase::Get();

define('C_AM', 1); // arithmetic mean
define('C_RMS', 2); // root mean square
define('C_WAM_SUM', 3); // weighted arithmetic mean by sum
define('C_TM', 4); // truncated mean
define('C_WAM_COUNT', 5); // weighted arithmetic mean by count
define('C_WAM_MAN1', 6); // weighted arithmetic mean by manual weights 1
define('C_WAM_PERF', 7); // weighted arithmetic mean by performance
define('C_WAM_AVG', 8); // weighted arithmetic mean by average
define('C_WAM_MAN2', 9); // weighted arithmetic mean by manual weights 2
define('C_WAM_MAN3', 10);// weighted arithmetic mean by manual weights 2
define('C_BAYES', 11); // nave Bayes
define('C_WRMS', 18); // weighted root mean square
/*
 * Methods operating on a limited set of descriptors.
 */
define('C_BAYES_CHOSEN', 12); // nave Bayes
define('C_AM_CHOSEN', 14); // arithmetic mean
define('C_RMS_CHOSEN', 15); // root mean square
define('C_TM_CHOSEN', 16); // truncated mean
define('C_MEDIAN_CHOSEN', 17); // median
define('C_WRMS_CHOSEN', 19); // weighted RMS

define('BAYES_PROB', 0.8);

define('CUTOFF_SCORE', 0.2);

define('FILES_IN_PASS', 25);


/**
 * Root mean square
 */
function calculate_rms($rgValues, $cCount)
{
    $dblSum = 0;
    foreach ($rgValues as $dblValue)
        $dblSum += ($dblValue * $dblValue);
    return sqrt($dblSum / (float) $cCount);
}

/**
 * Arithmetic mean
 */
function calculate_am($rgValues, $cCount)
{
    return array_sum($rgValues) / (float) $cCount;
}

/**
 * Weighted arithmetic mean
 */
function calculate_wam($rgValues, $rgWeight, $dblSumWeights)
{
    if ($dblSumWeights == 0) return 0;
    $dblSum = 0;
    foreach ($rgValues as $ix => $dbValue)
    {
        $dblSum += ($dbValue * $rgWeight[$ix]);
    }
    return $dblSum / $dblSumWeights;
}

/**
 * Weighted root mean square
 */
function calculate_wrms($rgValues, $rgWeight, $dblSumWeights)
{
    $dblSum = 0;
    foreach ($rgValues as $ix => $dblValue)
        $dblSum += $rgWeight[$ix] * ($dblValue * $dblValue);
    return sqrt($dblSum / (float) $dblSumWeights);
}

/**
 * Median
 */
function calculate_median($rgValues, $cCount)
{
    array_pad($rgValues, $cCount, 0);
    sort($rgValues);
    while (count($rgValues) > 2)
    {
        array_pop($rgValues);
        array_shift($rgValues);
    }
    if (count($rgValues) == 1)
        return array_pop($rgValues);
    else
        return array_sum($rgValues) / count($rgValues);
}

/**
 * Truncated arithmetic mean.
 */
function calculate_truncated_mean($rgValues, $cCount)
{
    // make sure 0-values are counted
    array_pad($rgValues, $cCount, 0);
    // sort
    sort($rgValues);
    // remove last and first
    array_pop($rgValues);
    array_shift($rgValues);
    return calculate_am($rgValues, $cCount - 2);
}

/**
 * Nave Bayes
 */
function calculate_bayes($rgValues)
{
    $dblProd = 1; $dblInvProd = 1;
    foreach ($rgValues as $dblValue)
    {
        $dblValue = max(0.01, min(0.99, $dblValue));
        $dblProd *= $dblValue;
        $dblInvProd *= (1 - $dblValue);
    }
    $dblProd /= BAYES_PROB;
    $dblInvProd /= (1 - BAYES_PROB);
    $dblSum = $dblProd + $dblInvProd;
    if ($dblSum != 0)
        return $dblProd / $dblSum;
    else
        return 0;
}

/**
 * Standard deviation
 */
function calculate_sd($dblMean, $rgValues, $cCount)
{
    $dblSum = 0;
    foreach ($rgValues as $ix => $dblValue)
    {
        $dblSum += ($dblValue - $dblMean) * ($dblValue - $dblMean);
    }
    return sqrt($dblSum / (float) $cCount);
}

/*
 * Which weights do we need?
 */
$rgWAMs = array(C_WAM_SUM, C_WAM_COUNT, C_WAM_AVG, C_WAM_MAN1, C_WAM_PERF, C_WAM_MAN2, C_WAM_MAN3);

/*
 * Let's figure out the weights for the weighted arithmetic means.
 */
$sqlSelectWeight = "SELECT ixType, dblWeightSum, dblWeightCount, dblWeightAvg, ".
  "dblWeightMan1, dblWeightMan2, dblWeightMan3, dblWeightPerformance ".
  "FROM tblType";
$rW =& $oDB->query($sqlSelectWeight);
$rgWeight = array();
$cCount = 0;
while ($r = $rW->fetchRow())
{
    $rgWeight[C_WAM_SUM][$r->ixType] = $r->dblWeightSum;
    $rgWeight[C_WAM_COUNT][$r->ixType] = $r->dblWeightCount;
    $rgWeight[C_WAM_MAN1][$r->ixType] = $r->dblWeightMan1;
    $rgWeight[C_WAM_MAN2][$r->ixType] = $r->dblWeightMan2;
    $rgWeight[C_WAM_MAN3][$r->ixType] = $r->dblWeightMan3;
    $rgWeight[C_WAM_PERF][$r->ixType] = $r->dblWeightPerformance;
    $rgWeight[C_WAM_AVG][$r->ixType] = $r->dblWeightAvg;
    $cCount++;
}
/*
 * Build an array of weight sums, indexed by which set of weights we are using
 * This is used later on for the weighted arithmetic mean, we'll just cache
 * this for later use.
 */
foreach ($rgWAMs as $ixWAM)
{
    $rgSumWeights[$ixWAM] = array_sum($rgWeight[$ixWAM]);
}
$rW->free();

/*
 * Determine the number of files. (Or more accurately, the range of ixFile)
 */
$cNumFiles = $oDB->extended->GetOne("SELECT MAX(ixFile) FROM tblFile");

/*
 * This section allows us to automatically start multiple subprocesses
 * of this script, to use more than one processor, and automatically
 * dividing the work load.
 * 
 * Examples:
 *   $ php calculate_summed_score.php 4 
 *     Runs 4 subprocesses. Output in output.txt.
 *   $ php calculate_summed_score.php
 *     Runs in 1 subprocess. Output to stdout.
 *   $ php calculate_summed_score.php 40 50 "one"
 *     Runs one process, calculating for files 40-90, using the
 *     process identifier "one".
 */
$ixRow = 0;
$sProcess = '';
if ($argc == 4 && is_numeric($argv[1]) && is_numeric($argv[2]))
{
    // use parameters as range to use
    $ixRow = (int) $argv[1];
    $cNumFiles = (int) $argv[2];
    $sProcess = $argv[3];
}
else if ($argc == 2 && is_numeric($argv[1]))
{
    // we are being asked to use multiple processes
    $cNumProc = (int) $argv[1];
    
    $cFilesPerProc = (int) ($cNumFiles / $cNumProc);
    // make sure we are a multiple of FILES_IN_PASS
    $cFilesPerProc += FILES_IN_PASS - ($cFilesPerProc % FILES_IN_PASS);
    
    exec('rm output.txt');
    
    $cmd = "php calculate_summed_score.php %d %d '%s' >> output.txt &";
    $n = 1;
    for ($i = 0; $i < $cNumFiles; $i += $cFilesPerProc)
    {
        $max = min($i + ($cFilesPerProc - 1), $cNumFiles);
        echo "Launching subprocess ($i -> $max)...\n";
        exec(sprintf($cmd, $i, $max, $n));
        $n++;
    }
    
    die();
}

echo "[$sProcess] Starting to sum ($cNumFiles files in total)...\n";

/*
 * Start calculating everything.
 */

$rgSelectedTypes = array(1,2,3,4,12,13,18,19);
$cSelectedCount = count($rgSelectedTypes);
while ($ixRow <= $cNumFiles)
{
    $sqlSelectScores = "SELECT ixFile1, ixFile2, ixType, dblScore ".
        "FROM tblScore WHERE ixFile1 BETWEEN $ixRow AND ".
        ($ixRow + FILES_IN_PASS - 1)." ORDER BY ixFile1, ixFile2";
    $res =& $oDB->query($sqlSelectScores);

    while ($row = $res->fetchRow())
    {
        $sKey = $row->ixFile1.'-'.$row->ixFile2;
        if ($sPrevKey != $sKey && !empty($sPrevKey))
        {
            // remember that $row has changed, so we can't look at those values
            list($ixFile1, $ixFile2) = split('-', $sPrevKey);
            
            /*
             * First run everything on the set of all descriptors
             */
            $dblRMS = calculate_rms($rgValues, $cCount);
            if ($dblRMS > CUTOFF_SCORE)
            {
                $dblRMS_SD = calculate_sd($dblRMS, $rgValues, $cCount);
                $rgInsert[] = '('.$ixFile1.', '.$ixFile2.', '.C_RMS.', '.$dblRMS.','.$dblRMS_SD.')';
            }
            
            $dblWRMS = calculate_wrms($rgValues, $rgWeight[C_WAM_MAN2], $rgSumWeights[C_WAM_MAN2]);
            if ($dblWRMS > CUTOFF_SCORE)
            {
                $dblWRMS_SD = calculate_sd($dblWRMS, $rgValues, $rgSumWeights[C_WAM_MAN2]);
                $rgInsert[] = '('.$ixFile1.', '.$ixFile2.', '.C_WRMS.', '.$dblWRMS.','.$dblWRMS_SD.')';
            }                

            $dblAM = calculate_am($rgValues, $cCount);
            if ($dblAM > CUTOFF_SCORE)
            {
                $dblAM_SD = calculate_sd($dblAM, $rgValues, $cCount);
                $rgInsert[] = '('.$ixFile1.', '.$ixFile2.', '.C_AM.', '.$dblAM.','.$dblAM_SD.')';
            }
            
            foreach ($rgWAMs as $ixWAM)
            {
                $dblWAM = calculate_wam($rgValues, $rgWeight[$ixWAM], $rgSumWeights[$ixWAM]);
                if ($dblWAM > CUTOFF_SCORE)
                {
                    $dblWAM_SD = calculate_sd($dblWAM, $rgValues, $rgSumWeights[$ixWAM]);
                    $rgInsert[] = '('.$ixFile1.', '.$ixFile2.', '.$ixWAM.', '.$dblWAM.','.$dblWAM_SD.')';
                }
            }
            
            $dblTM = calculate_truncated_mean($rgValues, $cCount);
            if ($dblTM > CUTOFF_SCORE)
            {
                $dblTM_SD = calculate_sd($dblTM, $rgValues, $cCount - 2);
                $rgInsert[] = '('.$ixFile1.', '.$ixFile2.', '.C_TM.', '.$dblTM.','.$dblTM_SD.')';
            }
            
            $dblBayes = calculate_bayes($rgValues)
            if ($dblBayes > CUTOFF_SCORE)
            {
                $rgInsert[] = '('.$ixFile1.', '.$ixFile2.', '.C_BAYES.', '.$dblBayes.',0)';
            }
            
            /*
             * Lets build a separate value array for the set of chosen descriptors.
             * This allows using the same functions to calculate summed scores.
             */
            $rgSelectedValues = array();
            foreach ($rgSelectedTypes as $ixType)
            {
                $rgSelectedValues[$ixType] = $rgValues[$ixType];
            }

            /*
             * Calculations for the chosen descriptors.
             */
            $dblMed = calculate_median($rgSelectedValues, $cSelectedCount);
            if ($dblMed > CUTOFF_SCORE)
            {
                $dblMed_SD = calculate_sd($dblMed, $rgSelectedValues, $cSelectedCount);
                $rgInsert[] = '('.$ixFile1.', '.$ixFile2.', '.C_MEDIAN_CHOSEN.', '.$dblMed.','.$dblMed_SD.')';
            }
            
            $dblRMS = calculate_rms($rgSelectedValues, $cSelectedCount);
            if ($dblRMS > CUTOFF_SCORE)
            {
                $dblRMS_SD = calculate_sd($dblRMS, $rgSelectedValues, $cSelectedCount);
                $rgInsert[] = '('.$ixFile1.', '.$ixFile2.', '.C_RMS_CHOSEN.', '.$dblRMS.','.$dblRMS_SD.')';
            }
            
            /*
             * Why is the WRMS here, yet not the WAM? Because the WAM was calculated using
             * all sets of weights, but we only chose one set of weights for the WRMS.
             * This means that when limiting the set of descriptors, yet using the same weights
             * we effectively get the third set of manual weights. Weird, yet correct.
             */
            $dblWRMS = calculate_wrms($rgSelectedValues, $rgWeight[C_WAM_MAN2], $rgSumWeights[C_WAM_MAN2]);
            if ($dblWRMS > CUTOFF_SCORE)
            {
                $dblWRMS_SD = calculate_sd($dblWRMS, $rgSelectedValues, $rgSumWeights[C_WAM_MAN2]);
                $rgInsert[] = '('.$ixFile1.', '.$ixFile2.', '.C_WRMS_CHOSEN.', '.$dblWRMS.','.$dblWRMS_SD.')';
            }                

            $dblAM = calculate_am($rgSelectedValues, $cSelectedCount);
            if ($dblAM > CUTOFF_SCORE)
            {
                $dblAM_SD = calculate_sd($dblAM, $rgSelectedValues, $cSelectedCount);
                $rgInsert[] = '('.$ixFile1.', '.$ixFile2.', '.C_AM_CHOSEN.', '.$dblAM.','.$dblAM_SD.')';
            }
            
            $dblTM = calculate_truncated_mean($rgSelectedValues, $cSelectedCount);
            if ($dblTM > CUTOFF_SCORE)
            {
                $dblTM_SD = calculate_sd($dblTM, $rgSelectedValues, $cSelectedCount - 2);
                $rgInsert[] = '('.$ixFile1.', '.$ixFile2.', '.C_TM_CHOSEN.', '.$dblTM.','.$dblTM_SD.')';
            }
            
            $dblBayes = calculate_bayes($rgSelectedValues)
            if ($dblBayes > CUTOFF_SCORE)
            {
                $rgInsert[] = '('.$ixFile1.', '.$ixFile2.', '.C_BAYES_CHOSEN.', '.$dblBayes.',0)';
            }                

            /*
             * Reset the value array for the next run.
             */
            unset($rgValues);
            $rgValues = array();
        }
        $rgValues[$row->ixType] = $row->dblScore;
        $sPrevKey = $sKey;
    }
    $res->free();
    
    /*
     * Insert all our cumulated results. Uses INSERT DELAYED since we don't need
     * to inspect this right away.
     */
    $sql = "INSERT DELAYED INTO tblSummedScore (ixFile1, ixFile2, ixScoreMethod, dblScore, dblSD) VALUES ";
    $sql.= implode(', ', $rgInsert);
    $oDB->query($sql);

    $i += count($rgInsert);
    $rgInsert = array();
    $ixRow += FILES_IN_PASS;
    echo "\t[".date('H:i:s')."] [$sProcess] Processed files starting from $ixRow ($i cumulative database inserts).\n";
}

echo "[$sProcess] Done!\n";

echo "[$sProcess] Entered $i scores into tblSummedScore!\n";
echo "[$sProcess] Finished!\n";
?>
